use crate::libvirt_lib_structures::nwfilter::NetworkFilterXML;
use crate::libvirt_lib_structures::XMLUuid;
use crate::libvirt_rest_structures::LibVirtStructError;
use std::net::{IpAddr, Ipv4Addr, Ipv6Addr};

#[derive(serde::Serialize, serde::Deserialize, Debug, Copy, Clone)]
pub enum NetworkFilterChainProtocol {
    Root,
    Mac,
    STP,
    VLAN,
    ARP,
    RARP,
    IPv4,
    IPv6,
}

impl NetworkFilterChainProtocol {
    pub fn from_xml(xml: &str) -> anyhow::Result<Self> {
        Ok(match xml {
            "root" => Self::Root,
            "mac" => Self::Mac,
            "stp" => Self::STP,
            "vlan" => Self::VLAN,
            "arp" => Self::ARP,
            "rarp" => Self::RARP,
            "ipv4" => Self::IPv4,
            "ipv6" => Self::IPv6,
            _ => {
                return Err(LibVirtStructError::ParseFilteringChain(format!(
                    "Unknown filtering chain: {xml}! "
                ))
                .into())
            }
        })
    }

    pub fn to_xml(&self) -> String {
        match self {
            Self::Root => "root",
            Self::Mac => "mac",
            Self::STP => "stp",
            Self::VLAN => "vlan",
            Self::ARP => "arp",
            Self::RARP => "rarp",
            Self::IPv4 => "ipv4",
            Self::IPv6 => "ipv6",
        }
        .to_string()
    }
}

#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
pub struct NetworkFilterChain {
    protocol: NetworkFilterChainProtocol,
    suffix: Option<String>,
}

impl NetworkFilterChain {
    pub fn from_xml(xml: &str) -> anyhow::Result<Option<Self>> {
        if xml.is_empty() {
            return Ok(None);
        }

        Ok(Some(match xml.split_once('-') {
            None => Self {
                protocol: NetworkFilterChainProtocol::from_xml(xml)?,
                suffix: None,
            },
            Some((prefix, suffix)) => Self {
                protocol: NetworkFilterChainProtocol::from_xml(prefix)?,
                suffix: Some(suffix.to_string()),
            },
        }))
    }

    pub fn to_xml(&self) -> String {
        match &self.suffix {
            None => self.protocol.to_xml(),
            Some(s) => format!("{}-{s}", self.protocol.to_xml()),
        }
    }
}

/// Network filter definition
#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
pub struct NetworkFilter {
    name: String,
    chain: Option<NetworkFilterChain>,
    priority: Option<i32>,
    uuid: Option<XMLUuid>,
    /// Referenced filters rules
    join_rules: Vec<String>,
    rules: Vec<NetworkFilterRule>,
}

impl NetworkFilter {
    pub fn from_xml(xml: NetworkFilterXML) -> anyhow::Result<Self> {
        let mut rules = Vec::with_capacity(xml.rules.len());
        for rule in &xml.rules {
            let mut selectors = Vec::new();

            // TODO : add other selectors

            rules.push(NetworkFilterRule {
                action: NetworkFilterAction::from_xml(&rule.action)?,
                direction: NetworkFilterDirection::from_xml(&rule.direction)?,
                priority: rule.priority,
                selectors,
            })
        }

        Ok(Self {
            name: xml.name,
            uuid: xml.uuid,
            chain: NetworkFilterChain::from_xml(&xml.chain)?,
            priority: xml.priority,
            join_rules: xml
                .filterrefs
                .iter()
                .map(|i| i.filter.to_string())
                .collect(),
            rules,
        })
    }
}

#[derive(serde::Serialize, serde::Deserialize, Debug, Copy, Clone)]
pub enum NetworkFilterAction {
    /// matching the rule silently discards the packet with no further analysis
    Drop,
    /// matching the rule generates an ICMP reject message with no further analysis
    Reject,
    /// matching the rule accepts the packet with no further analysis
    Accept,
    /// matching the rule passes this filter, but returns control to the calling filter for further
    /// analysis
    Return,
    /// matching the rule goes on to the next rule for further analysis
    Continue,
}

impl NetworkFilterAction {
    pub fn from_xml(xml: &str) -> anyhow::Result<Self> {
        Ok(match xml {
            "drop" => Self::Drop,
            "reject" => Self::Reject,
            "accept" => Self::Accept,
            "return" => Self::Return,
            "continue" => Self::Continue,
            s => {
                return Err(LibVirtStructError::ParseFilteringChain(format!(
                    "Unkown filter action {s}!"
                ))
                .into())
            }
        })
    }

    pub fn to_xml(&self) -> String {
        match self {
            Self::Drop => "drop",
            Self::Reject => "reject",
            Self::Accept => "accept",
            Self::Return => "return",
            Self::Continue => "continue",
        }
        .to_string()
    }
}

#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
pub enum NetworkFilterDirection {
    In,
    Out,
    InOut,
}

impl NetworkFilterDirection {
    pub fn from_xml(xml: &str) -> anyhow::Result<Self> {
        Ok(match xml {
            "in" => Self::In,
            "out" => Self::Out,
            "inout" => Self::InOut,
            s => {
                return Err(LibVirtStructError::ParseFilteringChain(format!(
                    "Unkown filter direction {s}!"
                ))
                .into())
            }
        })
    }

    pub fn to_xml(&self) -> String {
        match self {
            NetworkFilterDirection::In => "in",
            NetworkFilterDirection::Out => "out",
            NetworkFilterDirection::InOut => "inout",
        }
        .to_string()
    }
}

#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
pub enum Layer4State {
    NEW,
    ESTABLISHED,
    RELATED,
    INVALID,
    NONE,
}

#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
pub enum Layer4Type {
    TCP,
    UDP,
    SCTP,
    ICMP,
    TCPipv6,
    UDPipv6,
    SCTPipv6,
    ICMPipv6,
}

#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
pub struct NetworkFilterSelectorIP<IPv> {
    srcmacaddr: Option<String>,
    srcmacmask: Option<String>,
    dstmacaddr: Option<String>,
    dstmacmask: Option<String>,
    srcipaddr: Option<IPv>,
    srcipmask: Option<u8>,
    dstipaddr: Option<IPv>,
    dstipmask: Option<u8>,
    comment: Option<String>,
}

#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
pub enum NetworkFilterSelector {
    All,
    Mac {
        src_mac_addr: Option<String>,
        src_mac_mask: Option<String>,
        dst_mac_addr: Option<String>,
        dst_mac_mask: Option<String>,
        comment: Option<String>,
    },
    Arp {
        srcmacaddr: Option<String>,
        srcmacmask: Option<String>,
        dstmacaddr: Option<String>,
        dstmacmask: Option<String>,
        arpsrcipaddr: Option<IpAddr>,
        arpsrcipmask: Option<u8>,
        arpdstipaddr: Option<IpAddr>,
        arpdstipmask: Option<u8>,
        comment: Option<String>,
    },
    IPv4(NetworkFilterSelectorIP<Ipv4Addr>),
    IPv6(NetworkFilterSelectorIP<Ipv6Addr>),
    Layer4 {
        r#type: Layer4Type,
        srcmacaddr: Option<String>,
        srcipaddr: Option<IpAddr>,
        srcipmask: Option<u8>,
        dstipaddr: Option<IpAddr>,
        dstipmask: Option<u8>,
        /// Start of range of source IP address
        srcipfrom: Option<IpAddr>,
        /// End of range of source IP address
        srcipto: Option<IpAddr>,
        /// Start of range of destination IP address
        dstipfrom: Option<IpAddr>,
        /// End of range of destination IP address
        dstipto: Option<IpAddr>,
        srcportstart: Option<u16>,
        srcportend: Option<u16>,
        dstportstart: Option<u16>,
        dstportend: Option<u16>,
        state: Option<Layer4State>,
        comment: Option<String>,
    },
}

#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
pub struct NetworkFilterRule {
    action: NetworkFilterAction,
    direction: NetworkFilterDirection,
    /// optional; the priority of the rule controls the order in which the rule will be instantiated
    /// relative to other rules
    ///
    /// Valid values are in the range of -1000 to 1000.
    priority: Option<i32>,
    selectors: Vec<NetworkFilterSelector>,
}
